@@ -1,23 +1,11 @@
 
-#### Wavelets and Kernels ####
-## ---------------------------
-
-ricker <- function(n, a = 1)
-{
-	x <- seq(-n, n, length.out=n) / 2
-	C <- 2 / (sqrt(3 * a) * pi^(0.25))
-	m <- (1 - (x / a)^2)
-	g <- exp(-(x / a)^2 / 2)
-	C * m * g
-}
-
 #### Filtering and Smoothing ####
 ## -----------------------------
 
 filt1_ma <- function(x, width = 5L)
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( width %% 2L != 1L )
 		width <- 1 + 2 * (width %/% 2)
 	.Call(C_meanFilter, x, width, PACKAGE="matter")
@@ -26,16 +14,16 @@ filt1_ma <- function(x, width = 5L)
 filt1_conv <- function(x, weights)
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( length(weights) %% 2L != 1L )
-		stop("length of weights must be odd")
+		matter_error("length of weights must be odd")
 	.Call(C_linearFilter, x, weights, PACKAGE="matter")
 }
 
 filt1_gauss <- function(x, width = 5L, sd = (width %/% 2) / 2)
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( width %% 2L != 1L )
 		width <- 1L + 2L * as.integer(width %/% 2)
 	radius <- width %/% 2
@@ -48,7 +36,7 @@ filt1_bi <- function(x, width = 5L, sddist = (width %/% 2) / 2,
 	sdrange = mad(x, na.rm = TRUE))
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( width %% 2L != 1L )
 		width <- 1L + 2L * as.integer(width %/% 2)
 	.Call(C_bilateralFilter, x, width,
@@ -58,7 +46,7 @@ filt1_bi <- function(x, width = 5L, sddist = (width %/% 2) / 2,
 filt1_adapt <- function(x, width = 5L, spar = 1)
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( width %% 2L != 1L )
 		width <- 1L + 2L * as.integer(width %/% 2)
 	.Call(C_bilateralFilter, x, width,
@@ -69,11 +57,11 @@ filt1_diff <- function(x, niter = 3L, kappa = 50,
 	rate = 0.25, method = 1L)
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( kappa < 1 )
-		warning("kappa should be > 1")
+		matter_warn("kappa should be > 1")
 	if ( rate <= 0 || rate > 0.25 )
-		warning("rate should be positive and <= 0.25")
+		matter_warn("rate should be positive and <= 0.25")
 	.Call(C_diffusionFilter, x, niter,
 		kappa, rate, method, PACKAGE="matter")
 }
@@ -82,7 +70,7 @@ filt1_guide <- function(x, width = 5L, guide = x,
 	sdreg = mad(x, na.rm = TRUE))
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( width %% 2L != 1L )
 		width <- 1L + 2L * as.integer(width %/% 2)
 	if ( is.integer(x) && is.double(guide) )
@@ -97,7 +85,7 @@ filt1_pag <- function(x, width = 5L, guide = NULL,
 	sdreg = mad(x, na.rm = TRUE), ftol = 1/10)
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( width %% 2L != 1L )
 		width <- 1L + 2L * as.integer(width %/% 2)
 	if ( is.null(guide) )
@@ -114,11 +102,11 @@ filt1_sg <- function(x, width = 5L, order = min(3L, width - 2L),
 	deriv = 0, delta = 1)
 {
 	if ( !is.null(dim(x)) && length(dim(x)) != 1L )
-		stop("x must be a vector")
+		matter_error("x must be a vector")
 	if ( width %% 2L != 1L )
 		width <- 1L + 2L * as.integer(width %/% 2)
 	if ( width <= order )
-		stop("width must be larger than order")
+		matter_error("width must be larger than order")
 	b <- (1:width - (width %/% 2 + 1)) %*% t(rep.int(1, order + 1))
 	b <- b^(as.matrix(rep.int(1, width)) %*% (0:order))
 	weights <- pinv(b)[1 + deriv,]
@@ -150,22 +138,171 @@ filt1_fun <- function(method)
 	options[[match.arg(method, names(options))]]
 }
 
-convolve_at <- function(x, index, weights, ...)
+convolve_at <- function(x, index, weights,
+	margin = NULL, na.rm = FALSE)
 {
+	if ( is.null(margin) ) {
+		lenX <- length(x)
+	} else {
+		if ( !is.matrix(x) )
+			matter_error("x must be a vector or matrix")
+		lenX <- dim(x)[margin]
+	}
+	if ( is.matrix(index) )
+		index <- array2list(index, 1L)
 	if ( !is.list(index) )
-		stop("index must be a list")
-	if ( length(index) != length(x) )
-		stop("index and x must have the same length")
+		matter_error("index must be a list")
 	if ( is.numeric(weights) )
 		weights <- list(weights)
-	weights <- rep_len(weights, length(x))
+	weights <- rep_len(weights, lenX)
+	if ( length(index) != lenX )
+		matter_error("index and x must have the same extent")
+	if ( length(weights) != lenX )
+		matter_error("weights and x must have the same extent")
 	if ( !all(lengths(index) == lengths(weights)) )
-		stop("lengths of index and weights must match")
-	vapply(seq_along(x),
-		function(i) {
+		matter_error("lengths of index and weights must match")
+	if ( is.null(margin) ) {
+		FUN <- function(i) {
 			ii <- index[[i]]
-			sum(weights[[i]] * x[ii], ...)
-		}, numeric(1L))
+			sum(weights[[i]] * x[ii], na.rm=na.rm)
+		}
+		vapply(seq_along(x), FUN, numeric(1L))
+	} else {
+		FUN <- switch(margin,
+			function(i, w) colSums(w * x[i,,drop=FALSE], na.rm=na.rm),
+			function(i, w) colSums(w * t(x[,i,drop=FALSE]), na.rm=na.rm))
+		switch(margin,
+			t(mapply(FUN, index, weights)),
+			mapply(FUN, index, weights))
+	}
+}
+
+#### KNN Filtering and Smoothing ####
+## --------------------------------
+
+filtn_ma <- function(x, index, k = 5L, metric = "euclidean", p = 2)
+{
+	weights <- rep.int(1, k)
+	filtn_conv(x, index, weights, metric, p)
+}
+
+filtn_conv <- function(x, index, weights, metric = "euclidean", p = 2)
+{
+	if ( !is.numeric(weights) )
+		matter_error("weights must be numeric")
+	if ( missing(index) ) {
+		if ( is.null(dim(x)) ) {
+			index <- seq_along(x)
+		} else {
+			index <- expand.grid(lapply(dim(x), seq_len))
+		}
+	}
+	k <- length(weights)
+	nb <- knnsearch(index, k=k, metric=metric, p=p)
+	y <- convolve_at(x, nb, weights / sum(weights), na.rm=TRUE)
+	if ( !is.null(dim(x)) )
+		dim(y) <- dim(x)
+	y
+}
+
+filtn_gauss <- function(x, index, k = 5L, sd = median(r) / 2,
+	metric = "euclidean", p = 2)
+{
+	if ( missing(index) ) {
+		if ( is.null(dim(x)) ) {
+			index <- seq_along(x)
+		} else {
+			index <- expand.grid(lapply(dim(x), seq_len))
+		}
+	}
+	if ( !inherits(index, "kdtree") )
+		index <- kdtree(index)
+	nb <- knnsearch(index, k=k, metric=metric, p=p)
+	ds <- rowdist_at(index$data, ix=seq_along(x), iy=nb)
+	r <- vapply(ds, max, numeric(1L), na.rm=TRUE)
+	wts <- lapply(ds, function(d) exp(-d^2 / (2 * sd^2)))
+	wts <- lapply(wts, function(w) w / sum(w))
+	y <- convolve_at(x, nb, wts, na.rm=TRUE)
+	if ( !is.null(dim(x)) )
+		dim(y) <- dim(x)
+	y
+}
+
+filtn_bi <- function(x, index, k = 5L, sddist = median(r) / 2,
+	sdrange = mad(x, na.rm=TRUE), metric = "euclidean", p = 2)
+{
+	if ( missing(index) ) {
+		if ( is.null(dim(x)) ) {
+			index <- seq_along(x)
+		} else {
+			index <- expand.grid(lapply(dim(x), seq_len))
+		}
+	}
+	if ( !inherits(index, "kdtree") )
+		index <- kdtree(index)
+	nb <- knnsearch(index, k=k, metric=metric, p=p)
+	ds <- rowdist_at(index$data, ix=seq_along(x), iy=nb)
+	r <- vapply(ds, max, numeric(1L), na.rm=TRUE)
+	gwts <- lapply(ds,
+		function(d) exp(-d^2 / (2 * sddist^2)))
+	awts <- lapply(seq_along(x),
+		function(i) exp(-(x[nb[i,]] - x[i])^2 / (2 * sdrange^2)))
+	wts <- Map(function(g, a) g * a / sum(g * a), gwts, awts)
+	y <- convolve_at(x, nb, wts, na.rm=TRUE)
+	if ( !is.null(dim(x)) )
+		dim(y) <- dim(x)
+	y
+}
+
+filtn_adapt <- function(x, index, k = 5L, spar = 1,
+	metric = "euclidean", p = 2)
+{
+	if ( missing(index) ) {
+		if ( is.null(dim(x)) ) {
+			index <- seq_along(x)
+		} else {
+			index <- expand.grid(lapply(dim(x), seq_len))
+		}
+	}
+	if ( !inherits(index, "kdtree") )
+		index <- kdtree(index)
+	nb <- knnsearch(index, k=k, metric=metric, p=p)
+	ds <- rowdist_at(index$data, ix=seq_along(x), iy=nb)
+	r <- median(vapply(ds, max, numeric(1L), na.rm=TRUE))
+	xmedian <- median(x, na.rm=TRUE)
+	xmad <- mad(x, na.rm=TRUE)
+	xr <- max(x, na.rm=TRUE) - min(x, na.rm=TRUE)
+	dx <- lapply(seq_along(x),
+		function(i) abs(x[nb[i,,drop=TRUE]] - xmedian))
+	dx <- vapply(dx, mean, numeric(1L), na.rm=TRUE)
+	z <- abs(dx - xmad) / spar
+	sddist <- r * exp(-z) / sqrt(2)
+	sdrange <- xr * exp(-z) / sqrt(2)
+	gwts <- Map(function(d, sdd) exp(-d^2 / (2 * sdd^2)),
+		ds, sddist)
+	awts <- Map(function(i, sdr) exp(-(x[nb[i,]] - x[i])^2 / (2 * sdr^2)),
+		seq_along(x), sdrange)
+	wts <- Map(function(g, a) g * a / sum(g * a), gwts, awts)
+	y <- convolve_at(x, nb, wts, na.rm=TRUE)
+	if ( !is.null(dim(x)) )
+		dim(y) <- dim(x)
+	y
+}
+
+filtn_fun <- function(method)
+{
+	if ( is.character(method) )
+		method <- tolower(method)
+	options <- list(
+		"ma" = 			filtn_ma,
+		"mean" = 		filtn_ma,
+		"gauss" = 		filtn_gauss,
+		"gaussian" = 	filtn_gauss,
+		"bi" = 			filtn_bi,
+		"bilateral" = 	filtn_bi,
+		"adapt" = 		filtn_adapt,
+		"adaptive" = 	filtn_adapt)
+	options[[match.arg(method, names(options))]]
 }
 
 #### Alignment and warping ####
@@ -180,7 +317,7 @@ warp1_loc <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 	# find events in each signal
 	if ( missing(y) || is.null(y) ) {
 		if ( missing(ty) )
-			stop("either 'y' or 'ty' must be specified")
+			matter_error("either 'y' or 'ty' must be specified")
 		ly <- ty
 	} else {
 		ly <- NULL
@@ -227,7 +364,7 @@ warp1_loc <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 		}
 		tshift <- tout + shift
 	} else {
-		warning("no landmarks matched")
+		matter_warn("no landmarks matched")
 		tshift <- tout
 		path <- NULL
 	}
@@ -235,7 +372,7 @@ warp1_loc <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 	xout <- spline(tshift, x, xout=tout,
 		ties=list("ordered", mean))$y
 	attr(xout, "path") <- path
-	attr(xout, "tol") <- setNames(tol, tol.ref)
+	attr(xout, "tol") <- set_names(tol, tol.ref)
 	xout
 }
 
@@ -244,7 +381,7 @@ warp1_dtw <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 {
 	if ( missing(y) || is.null(y) ) {
 		if ( missing(ty) )
-			stop("either 'y' or 'ty' must be specified")
+			matter_error("either 'y' or 'ty' must be specified")
 		y <- rep_len(max(x, na.rm=TRUE), length(ty))
 		y <- simspec1(ty, y, tx, resolution=estres(tx, ref="x")^(-1))
 	}
@@ -266,7 +403,7 @@ warp1_dtw <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 	dn <- abs(reldiff(tx[length(tx)], ty[length(ty)], ref=tol.ref))
 	if ( tol < d0 || tol < dn ) {
 		tol <- max(d0, dn)
-		warning("'tol' must be greater than ", tol)
+		matter_warn("'tol' must be greater than ", tol)
 	}
 	path <- .Call(C_warpDTW, x, y, tx, ty,
 		tol, as_tol_ref(tol.ref), PACKAGE="matter")
@@ -278,7 +415,7 @@ warp1_dtw <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 		ties=list("ordered", mean), n=n)$y
 	xout <- approx(tx, x, xout=tout)$y
 	attr(xout, "path") <- path
-	attr(xout, "tol") <- setNames(tol, tol.ref)
+	attr(xout, "tol") <- set_names(tol, tol.ref)
 	xout
 }
 
@@ -288,7 +425,7 @@ warp1_cow <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 {
 	if ( missing(y) || is.null(y) ) {
 		if ( missing(ty) )
-			stop("either 'y' or 'ty' must be specified")
+			matter_error("either 'y' or 'ty' must be specified")
 		y <- rep_len(max(x, na.rm=TRUE), length(ty))
 		y <- simspec1(ty, y, tx, resolution=estres(tx, ref="x")^(-1))
 	}
@@ -312,17 +449,17 @@ warp1_cow <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 	}
 	# initialize nodes
 	if ( nbins < 2L )
-		stop("need at least 2 bins")
+		matter_error("need at least 2 bins")
 	xb <- findbins(x, nbins=nbins, dynamic=FALSE, limits.only=TRUE)
 	ix <- c(1L, xb$upper)
 	yb <- findbins(y, nbins=nbins, dynamic=FALSE, limits.only=TRUE)
 	iy <- c(1L, yb$upper)
 	if ( any(xb$size < 3L) || any(yb$size < 3L) )
-		stop("too many bins (need at least 3 samples per bin)")
+		matter_error("too many bins (need at least 3 samples per bin)")
 	dmax <- max(abs(reldiff(tx[ix], ty[iy], ref=tol.ref)))
 	if ( tol < dmax ) {
 		tol <- dmax
-		warning("'tol' must be greater than ", tol)
+		matter_warn("'tol' must be greater than ", tol)
 	}
 	# correlation optimized warping to align signals
 	path <- .Call(C_warpCOW, x, y, tx, ty,
@@ -337,7 +474,7 @@ warp1_cow <- function(x, y, tx = seq_along(x), ty = seq_along(y),
 		ties=list("ordered", mean), n=n)$y
 	xout <- approx(tx, x, xout=tout)$y
 	attr(xout, "path") <- path
-	attr(xout, "tol") <- setNames(tol, tol.ref)
+	attr(xout, "tol") <- set_names(tol, tol.ref)
 	xout
 }
 
@@ -351,7 +488,7 @@ icor <- function(x, y)
 		x <- x[!is.na(x)]
 	if ( anyNA(y) )
 		y <- y[!is.na(x)]
-	.Call(C_iCorr, x, y, PACKAGE="matter")
+	.Call(C_iCor, x, y, PACKAGE="matter")
 }
 
 warp1_fun <- function(method)
@@ -373,20 +510,24 @@ rescale_rms <- function(x, scale = 1)
 {
 	b <- sqrt(mean(x^2, na.rm=TRUE))
 	if ( b > 0 ) {
-		scale * x / b
+		y <- scale * x / b
 	} else {
-		rep.int(0, length(x))
+		y <- rep.int(0, length(x))
 	}
+	dim(y) <- dim(x)
+	y
 }
 
 rescale_sum <- function(x, scale = length(x))
 {
 	b <- sum(abs(x), na.rm=TRUE)
 	if ( b > 0 ) {
-		scale * x / b
+		y <- scale * x / b
 	} else {
-		rep.int(0, length(x))
+		y <- rep.int(0, length(x))
 	}
+	dim(y) <- dim(x)
+	y
 }
 
 rescale_ref <- function(x, ref = 1L, scale = 1, domain = NULL)
@@ -394,15 +535,17 @@ rescale_ref <- function(x, ref = 1L, scale = 1, domain = NULL)
 	if ( missing(domain) || is.null(domain) )
 		domain <- seq_along(x)
 	if ( length(x) != length(domain) )
-		stop("length of 'domain' must match length of 'x'")
+		matter_error("length of 'domain' must match length of 'x'")
 	if ( ref < min(domain) || ref > max(domain) )
-		stop("'ref' is outside of domain limits")
+		matter_error("'ref' is outside of domain limits")
 	i <- bsearch(ref, domain, nearest=TRUE)
 	if ( x[i] != 0 ) {
-		scale * x / x[i]
+		y <- scale * x / x[i]
 	} else {
-		rep.int(0, length(x))
+		y <- rep.int(0, length(x))
 	}
+	dim(y) <- dim(x)
+	y
 }
 
 rescale_range <- function(x, limits = c(0, 1))
@@ -443,17 +586,18 @@ rescale_fun <- function(method)
 #### Binning and downsampling ####
 ## -------------------------------
 
-binvec <- function(x, lower, upper, stat = "sum", prob = 0.5)
+roll <- function(x, width, na.drop = FALSE, fill = NA)
 {
-	if ( missing(lower) && missing(upper) ) {
-		stop("must specify one of 'lower' or 'upper'")
-	} else if ( missing(upper) ) {
-		upper <- c(lower[-1] - 1L, length(x))
-	} else if ( missing(lower) ) {
-		lower <- c(1L, upper[-length(upper)] + 1L)
-	}
-	.Call(C_binVector, x, as.integer(lower - 1L), as.integer(upper - 1L),
-		as_binstat(stat), prob, PACKAGE="matter")
+	r <- floor(width / 2)
+	x <- lapply(seq_along(x),
+		function(i) {
+			j <- (i - r):(i + r)
+			j[j < 1L | j > length(x)] <- NA
+			ifelse(!is.na(j), x[j], fill)
+		})
+	if ( na.drop )
+		x <- lapply(x, function(xi) xi[!is.na(xi)])
+	x
 }
 
 rollvec <- function(x, width, stat = "sum", prob = 0.5)
@@ -469,23 +613,47 @@ rollvec <- function(x, width, stat = "sum", prob = 0.5)
 		as_binstat(stat), prob, PACKAGE="matter")
 }
 
-findbins <- function(x, nbins, dynamic = TRUE,
-	niter = NA, limits.only = FALSE)
+binvec <- function(x, lower, upper, stat = "sum", prob = 0.5)
+{
+	if ( missing(lower) && missing(upper) ) {
+		matter_error("must specify one of 'lower' or 'upper'")
+	} else if ( missing(upper) ) {
+		upper <- c(lower[-1] - 1L, length(x))
+	} else if ( missing(lower) ) {
+		lower <- c(1L, upper[-length(upper)] + 1L)
+	}
+	.Call(C_binVector, x, as.integer(lower - 1L), as.integer(upper - 1L),
+		as_binstat(stat), prob, PACKAGE="matter")
+}
+
+findbins <- function(x, nbins, dynamic = FALSE, niter = NA,
+	overlap = 0, limits.only = FALSE)
 {
 	n <- floor(nbins)
-	width <- ceiling(length(x) / n)
-	breaks <- floor(seq(from=width, to=length(x) - width, length.out=n - 1L))
-	lower <- c(1L, breaks + 1L)
-	upper <- c(breaks, length(x))
-	if ( is.na(niter) ) {
-		niter <- floor(10 * length(x) / n)
-		check_converge <- TRUE
+	if ( overlap != 0 && !dynamic ) {
+		# overlapping intervals (based on graphics::co.intervals)
+		r <- length(x) / (nbins * (1 - overlap) + overlap)
+		breaks <- (0:(nbins - 1)) * (1 - overlap) * r
+		lower <- pmax(seq_along(x)[round(1 + breaks)], 0)
+		upper <- pmin(seq_along(x)[round(r + breaks)], length(x))
 	} else {
-		check_converge <- FALSE
+		# exclusive intervals
+		width <- ceiling(length(x) / n)
+		breaks <- floor(seq(from=width, to=length(x) - width, length.out=n - 1L))
+		lower <- c(1L, breaks + 1L)
+		upper <- c(breaks, length(x))
 	}
 	if ( dynamic ) {
+		if ( is.na(niter) ) {
+			niter <- floor(10 * length(x) / n)
+			check_converge <- TRUE
+		} else {
+			check_converge <- FALSE
+		}
 		if ( nbins < 3L )
-			stop("need >= 3 bins for dynamic binning")
+			matter_error("need >= 3 bins for dynamic binning")
+		if ( overlap != 0 )
+			matter_warn("ignoring overlap for dynamic binning")
 		# calculate SSE from linear regression in each bin
 		sse <- binvec(x, lower, upper, stat="sse")
 		trace <-  numeric(niter + 1L)
@@ -540,7 +708,7 @@ ltob <- function(x, t, lower, upper)
 	if ( is.double(x) && is.integer(t) )
 		t <- as.double(t)
 	if ( is.unsorted(t) )
-		stop("domain 't' must be sorted")
+		matter_error("domain 't' must be sorted")
 	lower <- as.integer(lower - 1L)
 	upper <- as.integer(upper - 1L)
 	.Call(C_downsampleLTOB, x, t, lower, upper, PACKAGE="matter")
@@ -553,7 +721,7 @@ lttb <- function(x, t, lower, upper)
 	if ( is.double(x) && is.integer(t) )
 		t <- as.double(t)
 	if ( is.unsorted(t) )
-		stop("domain 't' must be sorted")
+		matter_error("domain 't' must be sorted")
 	lower <- as.integer(lower - 1L)
 	upper <- as.integer(upper - 1L)
 	.Call(C_downsampleLTTB, x, t, lower, upper, PACKAGE="matter")
@@ -654,48 +822,108 @@ estbase_fun <- function(method)
 #### Noise estimation ####
 ## -----------------------
 
-estnoise_quant <- function(x, n = 25L, prob = 0.95, niter = 3L)
+estnoise_sd <- function(x, nbins = 1, overlap = 0.5,
+	k = 9, index = NULL)
 {
-	y <- filt1_diff(x, method=3L, niter=niter)
-	x <- abs(y - x)
-	width <- min(n, length(x))
-	rollvec(x, width, stat="quantile", prob=prob)
-}
-
-estnoise_diff <- function(x, nbins = 1L, dynamic = FALSE)
-{
-	fn <- function(xi) {
-		# mean absolute deviation of signal derivative
-		dx <- mean(diff(xi), na.rm=TRUE)
-		ns <- mean(abs(xi - dx), na.rm=TRUE)
-		noise <- rep.int(ns, length(xi))
-	}
-	if ( nbins > 1L ) {
-		xb <- findbins(x, nbins=nbins, dynamic=dynamic)
-		ns <- unlist(lapply(xb, fn))
-		noise <- lowess(ns)$y
+	if ( is.null(index) ) {
+		y <- filt1_gauss(x, width=k)
 	} else {
-		noise <- fn(x)
+		y <- filtn_gauss(x, index=index, k=k)
+	}
+	fn <- function(ei) sd(ei, na.rm=TRUE)
+	e <- abs(y - x)
+	if ( nbins > 1 ) {
+		e <- findbins(e, nbins=nbins, overlap=overlap)
+		i <- 0.5 * (attr(e, "lower") + attr(e, "upper"))
+		noise <- unlist(lapply(e, fn))
+		noise <- spline(i, noise, xout=seq_along(x))$y
+	} else {
+		noise <- rep.int(fn(e), length(x))
 	}
 	noise
 }
 
-estnoise_filt <- function(x, nbins = 1L, msnr = 2,
-	threshold = 0.5, peaks = FALSE)
+estnoise_mad <- function(x, nbins = 1, overlap = 0.5,
+	k = 9, index = NULL)
 {
-	if ( nbins > 1L ) {
-		# Gallia et al (2013) but with lowess
-		xb <- findbins(x, nbins=nbins, dynamic=FALSE)
-		noise <- lapply(xb, estnoise_filt, msnr=msnr,
-			threshold=threshold, peaks=peaks)
-		noise <- unlist(noise)
-		noise <- lowess(noise)$y
+	if ( is.null(index) ) {
+		y <- filt1_gauss(x, width=k)
 	} else {
+		y <- filtn_gauss(x, index=index, k=k)
+	}
+	fn <- function(ei) mad(ei, na.rm=TRUE)
+	e <- abs(y - x)
+	if ( nbins > 1 ) {
+		e <- findbins(e, nbins=nbins, overlap=overlap)
+		i <- 0.5 * (attr(e, "lower") + attr(e, "upper"))
+		noise <- unlist(lapply(e, fn))
+		noise <- spline(i, noise, xout=seq_along(x))$y
+	} else {
+		noise <- rep.int(fn(e), length(x))
+	}
+	noise
+}
+
+estnoise_quant <- function(x, nbins = 1, overlap = 0.5,
+	k = 9, prob = 0.95, index = NULL)
+{
+	if ( is.null(index) ) {
+		y <- filt1_gauss(x, width=k)
+	} else {
+		y <- filtn_gauss(x, index=index, k=k)
+	}
+	fn <- function(ei) quantile(ei, prob, na.rm=TRUE)
+	e <- abs(y - x)
+	if ( nbins > 1 ) {
+		e <- findbins(e, nbins=nbins, overlap=overlap)
+		i <- 0.5 * (attr(e, "lower") + attr(e, "upper"))
+		noise <- unlist(lapply(e, fn))
+		noise <- spline(i, noise, xout=seq_along(x))$y
+	} else {
+		noise <- rep.int(fn(e), length(x))
+	}
+	noise
+}
+
+estnoise_diff <- function(x, nbins = 1L, overlap = 0.5,
+	index = NULL)
+{
+	if ( is.null(index) ) {
+		dx <- diff(x)
+	} else {
+		dx <- (x - x[knnsearch(x, index, k=2)[,2L]])[-1L]
+	}
+	fn <- function(xi, dxi) {
+		# deviation between signal and its derivative
+		dxi <- mean(dxi, na.rm=TRUE)
+		mean(abs(xi - dxi), na.rm=TRUE)
+	}
+	if ( nbins > 1L ) {
+		xb <- findbins(x, nbins=nbins, overlap=overlap)
+		dxb <- findbins(dx, nbins=nbins, overlap=overlap)
+		i <- 0.5 * (attr(xb, "lower") + attr(xb, "upper"))
+		noise <- unlist(Map(fn, xb, dxb))
+		noise <- spline(i, noise, xout=seq_along(x))$y
+	} else {
+		noise <- rep.int(fn(x, dx), length(x))
+	}
+	noise
+}
+
+estnoise_filt <- function(x, nbins = 1L, overlap = 0.5,
+	msnr = 2, threshold = 0.5, isPeaks = FALSE, index = NULL)
+{
+	fn <- function(xi) {
 		# Xu and Freitas (2010) dynamic noise level
-		if ( isFALSE(peaks) || is.null(peaks) ) {
-			y <- sort(x[x > 0 & locmax(x)])
-		} else {
+		if ( isPeaks ) {
 			y <- sort(x[x > 0])
+		} else {
+			if ( is.null(index) ) {
+				peaks <- locmax(x)
+			} else {
+				peaks <- knnmax(x, index)
+			}
+			y <- sort(x[x > 0 & peaks])
 		}
 		if ( length(y) <= 1L )
 			return(rep.int(y, length(x)))
@@ -713,6 +941,16 @@ estnoise_filt <- function(x, nbins = 1L, msnr = 2,
 			# update linear model
 			fit <- lr_update(fit, i, y[i])
 		}
+		noise
+	}
+	if ( nbins > 1L ) {
+		# Gallia et al (2013) but with splines
+		xb <- findbins(x, nbins=nbins, overlap=overlap)
+		i <- 0.5 * (attr(xb, "lower") + attr(xb, "upper"))
+		noise <- unlist(lapply(xb, fn))
+		noise <- spline(i, noise, xout=seq_along(x))$y
+	} else {
+		noise <- fn(x)
 		noise <- rep.int(noise, length(x))
 	}
 	noise
@@ -751,35 +989,19 @@ lr_predict <- function(fit, xi) {
 	sum(fit$coef * c(1, xi))
 }
 
-estnoise_sd <- function(x, n = 25L, wavelet = ricker)
-{
-	if ( is.function(wavelet) )
-		x <- cwt(x, wavelet, scales=1)
-	width <- min(n, length(x))
-	rollvec(x, width, stat="sd")
-}
-
-estnoise_mad <- function(x, n = 25L, wavelet = ricker)
-{
-	if ( is.function(wavelet) )
-		x <- cwt(x, wavelet, scales=1)
-	width <- min(n, length(x))
-	rollvec(x, width, stat="sd")
-}
-
 estnoise_fun <- function(method)
 {
 	if ( is.character(method) )
 		method <- tolower(method)
 	options <- list(
+		"sd" = 			estnoise_sd,
+		"mad" = 		estnoise_mad,
 		"quant" = 		estnoise_quant,
 		"quantile" = 	estnoise_quant,
 		"diff" = 		estnoise_diff,
-		"derivate" = 	estnoise_diff,
+		"derivative" = 	estnoise_diff,
 		"filt" = 		estnoise_filt,
-		"filter" = 		estnoise_filt,
-		"sd" = 			estnoise_sd,
-		"mad" = 		estnoise_mad)
+		"filter" = 		estnoise_filt)
 	options[[match.arg(method, names(options))]]
 }
 
@@ -788,7 +1010,7 @@ estnoise_fun <- function(method)
 
 locmax <- function(x, width = 5L)
 {
-	.Call(C_localMaxima, x, as.integer(width), PACKAGE="matter")
+	y <- .Call(C_localMaxima, x, as.integer(width), PACKAGE="matter")
 }
 
 locmin <- function(x, width = 5L)
@@ -800,7 +1022,7 @@ findpeaks <- function(x, width = 5L, prominence = NULL,
 	snr = NULL, noise = "quant", bounds = TRUE,
 	relheight = 0.005, ...)
 {
-	peaks <- which(locmax(x, width))
+	peaks <- which(locmax(x, width=width))
 	ann <- data.frame(row.names=seq_along(peaks))
 	if ( isTRUE(bounds) )
 	{
@@ -854,7 +1076,7 @@ findpeaks <- function(x, width = 5L, prominence = NULL,
 		}
 	}
 	if ( length(ann) > 0L )
-		attributes(peaks) <- ann
+		peaks <- update_attr(peaks, ann)
 	peaks
 }
 
@@ -903,7 +1125,7 @@ findpeaks_cwt <- function(x, snr = 2, wavelet = ricker, scales = NULL,
 		ann$left_bounds <- bounds[[1L]] + 1L
 		ann$right_bounds <- bounds[[2L]] + 1L
 	}
-	attributes(peaks) <- ann
+	peaks <- update_attr(peaks, ann)
 	peaks
 }
 
@@ -984,6 +1206,15 @@ findridges <- function(x, maxdists, ngaps)
 	outridges
 }
 
+ricker <- function(n, a = 1)
+{
+	x <- seq(-n, n, length.out=n) / 2
+	C <- 2 / (sqrt(3 * a) * pi^(0.25))
+	m <- (1 - (x / a)^2)
+	g <- exp(-(x / a)^2 / 2)
+	C * m * g
+}
+
 cwt <- function(x, wavelet = ricker, scales = NULL)
 {
 	if ( is.null(scales) )
@@ -1020,7 +1251,7 @@ peakwidths <- function(x, peaks, domain = NULL,
 	if ( is.null(domain) )
 		domain <- seq_along(x)
 	if ( is.unsorted(domain) )
-		stop("'domain' must be sorted")
+		matter_error("'domain' must be sorted")
 	if ( ref == "height" )
 	{
 		# find peak boundaries if not provided
@@ -1061,6 +1292,7 @@ peakwidths <- function(x, peaks, domain = NULL,
 	ann$left_ips <- thresholds[[1L]]
 	ann$right_ips <- thresholds[[2L]]
 	widths <- ann$right_ips - ann$left_ips
+	widths <- update_attr(widths, ann)
 	attributes(widths) <- ann
 	widths
 }
@@ -1070,7 +1302,7 @@ peakareas <- function(x, peaks, domain = NULL)
 	if ( is.null(domain) )
 		domain <- seq_along(x)
 	if ( is.unsorted(domain) )
-		stop("'domain' must be sorted")
+		matter_error("'domain' must be sorted")
 	left_bounds <- attr(peaks, "left_bounds")
 	right_bounds <- attr(peaks, "right_bounds")
 	if ( is.null(left_bounds) || is.null(right_bounds) )
@@ -1117,7 +1349,7 @@ binpeaks <- function(peaklist, domain = NULL, xlist = peaklist,
 	tol = NA_real_, tol.ref = "abs", merge = FALSE, na.drop = TRUE)
 {
 	if ( any(lengths(peaklist) != lengths(xlist)) )
-		stop("lengths of 'peaklist' and 'xlist' must match")
+		matter_error("lengths of 'peaklist' and 'xlist' must match")
 	if ( is.na(tol) ) {
 		# guess tol as ~1/2 the min gap between same-signal peaks
 		ref <- ifelse(tol.ref == "abs", "abs", "y")
@@ -1135,7 +1367,7 @@ binpeaks <- function(peaklist, domain = NULL, xlist = peaklist,
 		}
 	}
 	if ( is.unsorted(domain) )
-		stop("'domain' must be sorted")
+		matter_error("'domain' must be sorted")
 	peaks <- numeric(length(domain))
 	x <- numeric(length(domain))
 	n <- numeric(length(domain))
@@ -1186,9 +1418,9 @@ mergepeaks <- function(peaks, n = nobs(peaks), x = peaks,
 	x <- as.vector(x)
 	peaks <- as.vector(peaks)
 	if ( length(peaks) != length(x) )
-		stop("length of 'peaks' and 'x' must match")
+		matter_error("length of 'peaks' and 'x' must match")
 	if ( is.unsorted(peaks, na.rm=TRUE) )
-		stop("'peaks' must be sorted")
+		matter_error("'peaks' must be sorted")
 	if ( is.na(tol) ) {
 		# guess tol as ~1/100 of average gap between peaks
 		ref <- ifelse(tol.ref == "abs", "abs", "y")
@@ -1266,40 +1498,48 @@ estres <- function(x, tol = NA, ref = NA_character_)
 	rx <- 2 * ((to / from ) - 1) / ((to / from) + 1)
 	rx <- ifelse(is.na(rx), Inf, rx)
 	dx <- diff(x)
-	if ( is.finite(tol) ) {
-		if ( (is.na(ref) && diff(range(rx)) > tol) || 
-			(!is.na(ref) && ref == "abs" ) )
-		{
-			dx <- dx[dx > tol]
-			if ( all(dx %% min(dx) <= tol) ) {
-				res <- c(absolute = min(dx))
+	dx <- dx[dx > .Machine$double.eps]
+	rx <- rx[rx > .Machine$double.eps]
+	if ( is.na(ref) ) {
+		if ( length(dx) && length(rx) ) {
+			if ( mad(dx / diff(range(x))) < mad(rx) ) {
+				ref <- "abs"
 			} else {
-				res <- c(absolute = NA_real_)
-			}
-		} else
-		{
-			rx <- rx[rx > tol]
-			if ( all(rx %% min(rx) <= tol) ) {
-				res <- c(relative = min(rx))
-			} else {
-				res <- c(relative = NA_real_)
-			}
-		}
-	} else {
-		if ( is.na(ref) ) {
-			if ( mad(dx) < mad(rx) ) {
-				res <- c(absolute = min(dx))
-			} else {
-				res <- c(relative = min(rx))
+				ref <- "x"
 			}
 		} else {
-			if ( ref == "abs" ) {
-				res <- c(absolute = min(dx))
+			if ( length(dx) ) {
+				ref <- "abs"
+			} else if ( length(rx) ) {
+				ref <- "x"
 			} else {
-				res <- c(relative = min(rx))
+				return(NA_real_)
 			}
 		}
 	}
+	if ( ref == "abs" ) {
+		if ( length(dx) ) {
+			res <- min(dx)
+		} else {
+			res <- Inf
+		}
+		if ( is.na(tol) || all(dx %% res <= tol) ) {
+			res <- c(absolute = res)
+		} else {
+			res <- c(absolute = NA_real_)
+		}
+	} else {
+		if ( length(dx) ) {
+			res <- min(rx)
+		} else {
+			res <- Inf
+		}
+		if ( is.na(tol) || all(rx %% res <= tol) ) {
+			res <- c(relative = res)
+		} else {
+			res <- c(relative = NA_real_)
+		}
+	}
 	res
 }
 
@@ -1336,7 +1576,7 @@ simspec <- function(n = 1L, npeaks = 50L,
 	baseline = 0, decay = 10, units = "relative")
 {
 	if ( length(x) != length(y) )
-		stop("length of 'x' and 'y' must match")
+		matter_error("length of 'x' and 'y' must match")
 	if ( length(domain) == 2L ) {
 		xout <- seq(from=domain[1L], to=domain[2L], length.out=size)
 	} else {
@@ -1377,7 +1617,7 @@ simspec <- function(n = 1L, npeaks = 50L,
 			peakwidths=peakwidths, sdnoise=sdnoise)
 		yout <- yout + b
 	}
-	structure(yout, domain=xout, peaks=x)
+	structure(yout, domain=xout, design=list(x=x, y=y))
 }
 
 simspec1 <- function(x, y, xout, peakwidths = NA_real_,
