@@ -6,13 +6,16 @@
 	options(
 		matter.compress.atoms = 3,
 		matter.default.nchunks = 20L,
+		matter.default.chunksize = NA_real_,
+		matter.default.serialize = NA,
 		matter.default.verbose = FALSE,
 		matter.matmul.bpparam = NULL,
 		matter.show.head = TRUE,
 		matter.show.head.n = 6L,
 		matter.coerce.altrep = FALSE,
 		matter.wrap.altrep = FALSE,
-		matter.dump.dir = tempdir(),
+		matter.temp.dir = tempdir(),
+		matter.temp.gc = TRUE,
 		matter.vizi.par = par_style_new(),
 		matter.vizi.panelgrid = NULL,
 		matter.vizi.trans3d = NULL,
@@ -22,45 +25,70 @@
 		matter.vizi.cpal = "Viridis")
 }
 
-#### Parallel RNG ####
-## --------------------
-
-RNGStreams <- function(n, parallel = FALSE) {
-	if ( isTRUE(parallel) )
-	{
-		if ( !"L'Ecuyer-CMRG" %in% RNGkind() )
-			RNGkind("L'Ecuyer-CMRG")
-		if ( missing(n) )
-			return(invisible())
-	}
-	seeds <- vector("list", n)
-	s <- getRNGStream()
-	if ( !is.null(s) )
-	{
-		if ( "L'Ecuyer-CMRG" %in% RNGkind() ) {
-			for ( i in seq_len(n) ) {
-				s <- nextRNGStream(s)
-				seeds[[i]] <- s
-			}
-		} else {
-			for ( i in seq_len(n) )
-				seeds[[i]] <- s
-		}
-	}
-	seeds
-}
-
-getRNGStream <- function(env = globalenv()) {
-	if ( exists(".Random.seed", envir=env) ) {
-		get(".Random.seed", envir=env)
+matter_defaults <- function(nchunks = 20L, chunksize = NA_real_,
+	serialize = NA, verbose = FALSE)
+{
+	if ( !missing(nchunks) ) {
+		nchunks <- as.integer(nchunks)
+		if ( !isTRUE(nchunks > 0L) || length(nchunks) != 1L )
+			matter_error("nchunks must be a single positive number")
+		options(matter.default.nchunks=nchunks)
 	} else {
-		NULL
+		nchunks <- getOption("matter.default.nchunks")
+	}
+	if ( !missing(chunksize) ) {
+		if ( isFALSE(chunksize > 0L) || length(chunksize) != 1L )
+			matter_error("chunksize must be a single positive number or NA")
+		options(matter.default.chunksize=as_size(chunksize))
+	} else {
+		chunksize <- getOption("matter.default.chunksize")
+	}
+	if ( !missing(serialize) ) {
+		if ( !is.logical(serialize) || length(serialize) != 1L )
+			matter_error("serialize must be a single logical value")
+		options(matter.default.serialize=serialize)
+	} else {
+		serialize <- getOption("matter.default.serialize")
+	}
+	if ( !missing(verbose) ) {
+		if ( !isTRUE(verbose) && !isFALSE(verbose) )
+			matter_error("verbose must be TRUE or FALSE")
+		options(matter.default.verbose=verbose)
+	} else {
+		verbose <- getOption("matter.default.verbose")
+	}
+	defaults <- list(nchunks=nchunks, chunksize=chunksize,
+		serialize=serialize, verbose=verbose)
+	if ( nargs() > 0L ) {
+		invisible(defaults)
+	} else {
+		defaults
 	}
 }
 
-setRNGStream <- function(seed = NULL, env = globalenv()) {
-	if ( !is.null(seed) && is.integer(seed) )
-		assign(".Random.seed", seed, envir=env)
+#### Environments and closures ####
+## ---------------------------------
+
+matter_ns <- environment(NULL)
+
+matter_env <- function() matter_ns
+
+copy_env <- function(envir, parent = baseenv()) {
+	copy <- new.env(parent=parent)
+	for ( name in ls(envir=envir) )
+		assign(name, get(name, envir=envir), envir=copy)
+	copy
+}
+
+isofun <- function(fun, envir = baseenv()) {
+	environment(fun) <- envir
+	fun
+}
+
+isoclos <- function(fun, envir = baseenv()) {
+	enclos <- copy_env(environment(fun), parent=envir)
+	environment(fun) <- enclos
+	fun
 }
 
 #### Normalize subscripts ####
@@ -131,7 +159,7 @@ linear_ind <- function(index, dim, rowMaj = FALSE) {
 	for ( j in seq_along(dim) ) {
 		i <- index[,j]
 		if ( any(!is.na(i) & (i <= 0 | i > dim[j])) )
-			stop("subscript out of bounds")
+			matter_error("subscript out of bounds")
 	}
 	if ( rowMaj ) {
 		strides <- c(rev(cumprod(rev(dim[-1L]))), 1L)
@@ -382,9 +410,12 @@ as_dist <- function(x) {
 	make_code(codes, x[1L], nomatch=1L)
 }
 
-as_weights <- function(x) {
-	codes <- c("gaussian", "adaptive")
-	make_code(codes, x[1L], nomatch=1L)
+as_size <- function(x) {
+	if ( is.null(names(x)) ) {
+		x[1L]
+	} else {
+		size_bytes(unname(x[1L]), units=names(x[1L]))
+	}
 }
 
 #### Data structure utility functions ####
@@ -394,7 +425,7 @@ as_real_memory_matrix <- function(x) {
 	if ( is.matrix(x) || is(x, "Matrix") ) {
 		x
 	} else {
-		warning("coercing input to a local matrix")
+		matter_warn("coercing input to a local matrix")
 		as.matrix(x)
 	}
 }
@@ -409,13 +440,13 @@ I <- function(x) {
 
 non_null <- function(x) !vapply(x, is.null, logical(1L))
 
-is_nil <- function(x) is.na(x) || is.null(x)
+is_null_or_na <- function(x) is.null(x) || is.na(x)
 
 is_discrete <- function(x) {
 	is.factor(x) || is.character(x) || is.logical(x)
 }
 
-set_attr <- function(x, attr) {
+update_attr <- function(x, attr) {
 	for ( nm in names(attr) )
 		attr(x, nm) <- attr[[nm]]
 	x
@@ -443,8 +474,8 @@ combine_names <- function(x1, x2) {
 	nm2 <- names(x2)
 	if ( is.null(nm1) && is.null(nm2) )
 		return(names(x1))
-	nm1 <- if (is.null(nm1)) character(length(x1)) else nm1
-	nm2 <- if (is.null(nm2)) character(length(x2)) else nm2
+	nm1 <- nm1 %||% character(length(x1))
+	nm2 <- nm2 %||% character(length(x2))
 	c(nm1, nm2)
 }
 
@@ -478,8 +509,8 @@ cbind_dimnames <- function(x1, x2) {
 	c2 <- colnames(x2)
 	if ( is.null(c1) && is.null(c2) )
 		return(dimnames(x1))
-	c1 <- if (is.null(c1)) character(ncol(x1)) else c1
-	c2 <- if (is.null(c2)) character(ncol(x2)) else c2
+	c1 <- c1 %||% character(ncol(x1))
+	c2 <- c2 %||% character(ncol(x2))
 	list(rownames(x1), c(c1, c2))
 }
 
@@ -488,8 +519,8 @@ rbind_dimnames <- function(x1, x2) {
 	r2 <- rownames(x2)
 	if ( is.null(r1) && is.null(r2) )
 		return(dimnames(x1))
-	r1 <- if (is.null(r1)) character(nrow(x1)) else r1
-	r2 <- if (is.null(r2)) character(nrow(x2)) else r2
+	r1 <- r1 %||% character(nrow(x1))
+	r2 <- r2 %||% character(nrow(x2))
 	list(c(r1, r2), colnames(x1))
 }
 
@@ -567,6 +598,10 @@ normalize_lengths2 <- function(list1, list2) {
 	list(list1, list2)
 }
 
+array2list <- function(X, MARGIN) {
+	apply(X, MARGIN, identity, simplify=FALSE)
+}
+
 nlines <- function(x) {
 	xsub <- gsub("\n", "", x, fixed=TRUE)
 	nchar(x) - nchar(xsub) + 1L
@@ -576,13 +611,17 @@ nlines <- function(x) {
 ## -----------------------------
 
 show_matter_mem <- function(x) {
-	rmem <- size_bytes(object.size(x))
-	rmem <- format(rmem, units="auto")
+	size <- size_bytes(object.size(x))
+	size <- format(size, units="auto")
 	if ( is.matter(x) ) {
-		vmem <- format(vm_used(x), units="auto")
-		cat("(", rmem, " real", " | ", vmem, " virtual)\n", sep="")
+		shm <- format(shm_used(x), units="auto")
+		vm <- format(vm_used(x), units="auto")
+		cat("(",
+			size, " real", " | ",
+			shm, " shared", " | ",
+			vm, " virtual)\n", sep="")
 	} else {
-		cat("(", rmem, " real)\n", sep="")
+		cat("(", size, " real)\n", sep="")
 	}
 }
 
@@ -706,18 +745,43 @@ preview_Nd_array <- function(x, n = getOption("matter.show.head.n"), ...) {
 preview_list <- function(x, n = getOption("matter.show.head.n"), ...) {
 	n1 <- min(n, length(x))
 	for ( i in 1:n1 ) {
-		fmt <- preview_vector_data(x[[i]], n, ...)
-		if ( !is.null(names(x)) ) {
-			rownames(fmt) <- paste0("$", names(x)[i])
+		xi <- x[[i]]
+		if ( is.null(dim(xi)) ) {
+			fmt <- preview_vector_data(xi, n, ...)
 		} else {
-			rownames(fmt) <- paste0("[[", i, "]]")
+			fmt <- preview_matrix_data(xi, n, ...)
 		}
+		if ( !is.null(names(x)) ) {
+			nm <- paste0("$", names(x)[i])
+		} else {
+			nm <- paste0("[[", i, "]]")
+		}
+		if ( nrow(fmt) > 1L )
+			nm <- c(nm, character(nrow(fmt) - 1L))
+		rownames(fmt) <- nm
 		print(fmt, quote=FALSE, right=TRUE)
 	}
 	if ( length(x) > n1 )
 		cat("...\n")
 }
 
+preview_recursive <- function(x, n = getOption("matter.show.head.n"), ...) {
+	n1 <- min(n, length(x))
+	for ( i in 1:n1 ) {
+		xi <- x[[i]]
+		if ( !is.null(names(x)) ) {
+			cat("$", names(x)[i], "\n", sep="")
+		} else {
+			cat("[[", i, "]]\n", sep="")
+		}
+		preview_list(xi)
+		if ( i < n1 )
+			cat("\n")
+	}
+	if ( length(x) > n1 )
+		cat("...\n")
+}
+
 preview_table <- function(x, n = getOption("matter.show.head.n"), cls = NULL, ...) {
 	more_i <- nrow(x) > n
 	more_j <- ncol(x) > n
@@ -771,7 +835,7 @@ preview_table <- function(x, n = getOption("matter.show.head.n"), cls = NULL, ..
 apply_int <- function(X, MARGIN, FUN, FUN.VALUE, ...) {
 	FUN <- match.fun(FUN)
 	if ( !MARGIN %in% c(1L, 2L) )
-		stop("MARGIN must be 1 or 2")
+		matter_error("MARGIN must be 1 or 2")
 	switch(MARGIN,
 		vapply(seq_len(nrow(X)), function(i)
 			FUN(X[i,,drop=TRUE], ...), FUN.VALUE),
@@ -780,26 +844,16 @@ apply_int <- function(X, MARGIN, FUN, FUN.VALUE, ...) {
 }
 
 bplapply_int <- function(X, FUN, ..., BPPARAM = NULL) {
-	if ( !is.null(BPPARAM) ) {
-		bplapply(X, FUN, ..., BPPARAM=BPPARAM)
+	if ( is.null(BPPARAM) ) {
+		ans <- vector("list", length=length(X))
+		for ( i in seq_along(X) )
+			ans[[i]] <- FUN(X[[i]], ...)
+		set_names(ans, names(X))
 	} else {
-		lapply(X, FUN, ...)
+		bplapply(X, FUN, ..., BPPARAM=BPPARAM)
 	}
 }
 
-roll <- function(x, width, na.drop = FALSE, fill = NA) {
-	r <- floor(width / 2)
-	x <- lapply(seq_along(x),
-		function(i) {
-			j <- (i - r):(i + r)
-			j[j < 1L | j > length(x)] <- NA
-			ifelse(!is.na(j), x[j], fill)
-		})
-	if ( na.drop )
-		x <- lapply(x, function(xi) xi[!is.na(xi)])
-	x
-}
-
 # A sequence with half-bin-widths in relative units
 # x = bin center, y = half-width, d = relative diff
 # y[n] = d * x[n]
@@ -828,7 +882,7 @@ shingles <- function(x, breaks, overlap = 0.5, labels = NULL)
 		labeller <- function(b) paste0("[", b[1L], ",", b[2L], "]")
 		labels <- apply(breaks, 1L, labeller)
 	} else if ( length(labels) != length(y) ) {
-		stop("length of labels not equal to length of breaks")
+		matter_error("length of labels not equal to length of breaks")
 	}
 	attr(y, "breaks") <- breaks
 	attr(y, "counts") <- lengths(y)
@@ -841,7 +895,8 @@ shingles <- function(x, breaks, overlap = 0.5, labels = NULL)
 avg <- function(x, center = mean)
 {
 	x <- x[!is.na(x)]
-	if ( is.numeric(x) ) {
+	if ( is.numeric(x) || is.complex(x) ) {
+		center <- match.fun(mean)
 		y <- center(x)
 	} else {
 		ux <- unique(x)
@@ -888,7 +943,7 @@ predict_class <- function(scores) {
 }
 
 # matrix pseudoinverse based on MASS::ginv
-pinv <- function (x, tol = sqrt(.Machine$double.eps)) 
+pinv <- function(x, tol = sqrt(.Machine$double.eps))
 {
 	x <- as.matrix(x)
 	sv <- svd(x)
@@ -902,6 +957,22 @@ pinv <- function (x, tol = sqrt(.Machine$double.eps))
 	}
 }
 
+# fix file paths
+fix_path <- function(path, winslash = "\\", mustWork = NA) {
+	file <- !is_shared_memory_pattern(path)
+	if ( any(file) )
+		path[file] <- normalizePath(path[file],
+			winslash=winslash, mustWork=mustWork)
+	path
+}
+
+# check file paths
+same_path <- function(path1, path2) {
+	path1 <- fix_path(path1, mustWork=FALSE)
+	path2 <- fix_path(path2, mustWork=FALSE)
+	identical(path1, path2)
+}
+
 #### Utilities for raw bytes and memory ####
 ## -----------------------------------------
 
@@ -937,17 +1008,21 @@ raw2hex <- function(x, uppercase = FALSE) {
 }
 
 # create a uuid
-
-uuid <- function(uppercase = FALSE) {
+uuid <- function(uppercase = FALSE)
+{
+	oseed <- getRNGStream()
+	on.exit(setRNGStream(oseed))
+	setRNGStream(get_shared_RNGStream())
 	hex <- as.raw(0:255)
 	version <- hex[65:80] # 0100 xxxx (version 4)
 	variant <- hex[129:192] # 10xx xxxx (variant 1)
-	time_low <- sample(hex, 4)
-	time_hi <- sample(hex, 2)
-	time_hi_and_version <- c(sample(version, 1), sample(hex, 1))
-	clock_seq_hi_and_res <- sample(variant, 1)
-	clock_seq_low <- sample(hex, 1)
-	node <- sample(hex, 6)
+	time_low <- sample(hex, 4L, TRUE)
+	time_hi <- sample(hex, 2L, TRUE)
+	time_hi_and_version <- c(sample(version, 1L), sample(hex, 1L))
+	clock_seq_hi_and_res <- sample(variant, 1L)
+	clock_seq_low <- sample(hex, 1L)
+	node <- sample(hex, 6L, TRUE)
+	set_shared_RNGStream(getRNGStream()$seed)
 	bytes <- c(time_low, time_hi, time_hi_and_version,
 		clock_seq_hi_and_res, clock_seq_low, node)
 	string <- c(
@@ -967,7 +1042,19 @@ uuid <- function(uppercase = FALSE) {
 
 # creates internal S3 class 'size_bytes'
 # (similar to 'object_size' but works w/ vectors)
-size_bytes <- function(x) {
+size_bytes <- function(x, units = "B") {
+	if ( inherits(x, "size_bytes") )
+		return(x)
+	units <- match.arg(units,
+		c("bytes", "B", "KB", "MB", "GB", "TB", "PB"))
+	x <- set_names(as.numeric(x), names(x))
+	x <- switch(units,
+		"bytes" = , "B" = x,
+		"KB" = x * 1000,
+		"MB" = x * 1000^2,
+		"GB" = x * 1000^3,
+		"TB" = x * 1000^4,
+		"PB" = x * 1000^5)
 	class(x) <- "size_bytes"
 	x
 }
@@ -975,12 +1062,18 @@ size_bytes <- function(x) {
 # based on utils:::format.object_size
 format.size_bytes <- function(x, units = "auto", ...)
 {
-	units <- match.arg(units, c("auto",
-		"B", "KB", "MB", "GB", "TB", "PB"))
-	mx <- max(x, na.rm=TRUE)
+	units <- match.arg(units,
+		c("auto", "B", "KB", "MB", "GB", "TB", "PB"))
+	if ( all(is.na(x)) ) {
+		mx <- NA_real_
+	} else {
+		mx <- max(x, na.rm=TRUE)
+	}
 	if ( units == "auto" )
 		units <- if ( is.na(mx) )
 			" "
+		else if ( mx >= 1000^5 )
+			"PB"
 		else if ( mx >= 1000^4 )
 			"TB"
 		else if ( mx >= 1000^3 )
@@ -992,11 +1085,11 @@ format.size_bytes <- function(x, units = "auto", ...)
 		else "B"
 	sizes <- switch(units,
 		" " = , "B" = x,
-		"KB" = round(x/1000, 2L),
-		"MB" = round(x/1000^2, 2L),
-		"GB" = round(x/1000^3, 2L),
-		"TB" = round(x/1000^4, 2L),
-		"PB" = round(x/1000^5, 2L))
+		"KB" = roundup(x / 1000, 2L),
+		"MB" = roundup(x / 1000^2, 2L),
+		"GB" = roundup(x / 1000^3, 2L),
+		"TB" = roundup(x / 1000^4, 2L),
+		"PB" = roundup(x / 1000^5, 2L))
 	label <- switch(units, "B"="bytes", units)
 	set_names(paste(sizes, label), names(x))
 }
@@ -1013,50 +1106,160 @@ print.size_bytes <- function(x, units = "auto",
 	structure(NextMethod(), class="size_bytes")
 }
 
-# based on pryr::mem_used and pryr::mem_change
+roundup <- function(x, digits = 0) {
+	ceiling(10^digits * x) / 10^(digits)
+}
+
 mem <- function(x, reset = FALSE)
 {
-	if ( !missing(x) ) {
-		rmem <- as.numeric(object.size(x))
-		vmem <- as.numeric(vm_used(x))
-		mem <- c("real"=rmem, "virtual"=vmem)
-	} else {
+	if ( missing(x) ) {
 		cell.size <- c(Ncells=56, Vcells=8)
 		gc.result <- gc(reset=reset)
-		gc.cols <- c(1L, 3L, ncol(gc.result) - 1L)
-		mem <- colSums(gc.result[,gc.cols] * cell.size)
-		names(mem) <- c("used", "gc", "max")
+		gc.cols <- c(1L, ncol(gc.result) - 1L)
+		real <- unname(colSums(gc.result[,gc.cols] * cell.size))
+		temp <- sum(unname(sizeof_file_resource(owned=TRUE)))
+		shm.used <- sum(unname(sizeof_memory_resource(owned=TRUE)))
+		if ( reset ) {
+			set_shared_memory_max(shm.used)
+			set_shared_file_max(temp)
+			shm.max <- shm.used
+		} else {
+			shm.max <- update_shared_memory_max()
+		}
+		shared <- unname(c(shm.used, shm.max))
+		mem <- c(
+			"real"=real[1L],
+			"shared"=shared[1L],
+			"max real"=real[2L],
+			"max shared"=shared[2L],
+			"temp"=temp)
+	} else {
+		size <- as.numeric(object.size(x))
+		shm <- as.numeric(shm_used(x))
+		vm <- as.numeric(vm_used(x))
+		mem <- c("real"=size, "shared"=shm, "virtual"=vm)
 	}
 	size_bytes(mem)
 }
 
-memtime <- function(expr)
+memcl <- function(cl = bpparam(), reset = FALSE)
 {
-	start <- mem(reset = TRUE)
-	t.start <- proc.time()
+	if ( is(cl, "SnowParam") && !is(cl, "MulticoreParam") )
+		cl <- bpbackend(cl)
+	if ( !inherits(cl, "cluster") )
+		matter_error("cannot monitor memory for cluster of class: ",
+			sQuote(class(cl)))
+	if ( inherits(cl, "NULLcluster") ) {
+		matter_warn("cluster is NULL; has it been started yet?")
+		return(data.frame())
+	}
+	if ( reset ) {
+		ans <- clusterEvalQ(cl, matter::mem(reset=TRUE))
+	} else {
+		ans <- clusterEvalQ(cl, matter::mem(reset=FALSE))
+	}
+	node <- vapply(cl, function(worker) worker$host, character(1L))
+	ans <- do.call(rbind, ans)
+	ans <- apply(ans, 2L, size_bytes, simplify=FALSE)
+	ans <- c(list(node=node), ans)
+	data.frame(ans, check.names=FALSE)
+}
+
+memtime <- function(expr, verbose = NA, BPPARAM = NULL)
+{
+	if ( is.na(verbose) )
+		verbose <- getOption("matter.default.verbose")
 	expr <- substitute(expr)
+	expr_string <- paste0(deparse(expr), collapse="\n")
+	mem.cols <- seq_len(4L)
+	start <- mem(reset=TRUE)[mem.cols]
+	matter_log(tstamp(), verbose=verbose)
+	if ( !is.null(BPPARAM) )
+	{
+		if ( !bpisup(BPPARAM) ) {
+			managed.cl <- TRUE
+			matter_log("Starting cluster...", verbose=verbose)
+			bpstart(BPPARAM)
+		} else {
+			managed.cl <- FALSE
+		}
+		if ( inherits(BPPARAM, "SnowParam") ) {
+			monitored.cl <- TRUE
+			start.cl <- memcl(BPPARAM, reset=TRUE)
+		} else {
+			monitored.cl <- FALSE
+		}
+	}
+	matter_log("Timing started at ", format(Sys.time(), "%X"),
+		verbose=verbose)
+	matter_log(expr_string, verbose=verbose)
+	t.start <- proc.time()
 	eval(expr, parent.frame())
 	rm(expr)
 	t.end <- proc.time()
-	end <- mem(reset = FALSE)
-	mem <- c(start[1], end[1], end[3])
-	mem <- c(format(size_bytes(mem)),
-		format(size_bytes(end[3] - end[1])),
-		paste0(round(t.end[3] - t.start[3], 4L), " sec"))
-	names(mem) <- c("start", "finish",
-		"max", "overhead", "time")
-	print.default(mem, quote=FALSE, right=TRUE)
+	matter_log("Timing ended at ", format(Sys.time(), "%X"),
+		verbose=verbose)
+	if ( !is.null(BPPARAM) )
+	{
+		if ( monitored.cl ) {
+			matter_log("Collecting cluster memory use...", verbose=verbose)
+			end.cl <- memcl(BPPARAM, reset=FALSE)
+		}
+		if ( managed.cl ) {
+			matter_log("Stopping cluster...", verbose=verbose)
+			bpstop(BPPARAM)
+		}
+	}
+	matter_log(tstamp(), verbose=verbose)
+	end <- mem(reset=FALSE)[mem.cols]
+	change <- c(
+		"real"=unname(end["real"] - start["real"]),
+		"shared"=unname(end["shared"] - start["shared"]))
+	result <- list(
+		"start"=start, "end"=end,
+		"change"=size_bytes(change))
+	overhead <- c(
+		"real"=unname(end["max real"] - end["real"]),
+		"shared"=unname(end["max shared"] - end["shared"]))
+	total <- unname(sum(end[["max real"]]) + sum(end[["max shared"]]))
+	if ( !is.null(BPPARAM) )
+	{
+		if ( monitored.cl ) {
+			total.cl <- sum(end.cl[["max real"]])
+			result$cluster <- list(
+				"start"=start.cl, "end"=end.cl,
+				"total"=size_bytes(total.cl))
+			overhead["real"] <- overhead["real"] + total.cl
+			total <- total + total.cl
+		}
+	}
+	result$overhead <- size_bytes(overhead)
+	result$total <- size_bytes(total)
+	result$time <- t.end - t.start
+	structure(result, class="memtime")
+}
+
+tstamp <- function(pre = "##------ ", post = " ------##") {
+	paste0(pre, date(), post)
 }
 
 profmem <- function(expr)
 {
-	.Deprecated("memtime")
-	memtime(expr)
+	.Defunct("memtime")
 }
 
 #### Formula parsing ####
 ## ----------------------
 
+iQuote <- function(x, q = "`")
+{
+	if ( length(x) > 0 ) {
+		paste0(q, as.character(x), q)
+	} else {
+		character(0L)
+	}
+}
+
 parse_formula <- function(formula, envir = NULL, eval = !missing(envir))
 {
 	e <- environment(formula)
@@ -1129,3 +1332,71 @@ parse_side <- function(formula, envir = NULL, eval = FALSE)
 	side
 }
 
+eval_exprs <- function(exprs, data, i = NULL, j = NULL,
+	split_along = NULL, group = NULL, reduce = "+",
+	recursive = !is.null(split_along))
+{
+	ans <- vector("list", length=length(exprs))
+	for ( k in seq_along(exprs) )
+	{
+		ans[[k]] <- eval_at(exprs[[k]], data=data, i=i, j=j,
+			split_along=split_along, group=group, reduce=reduce,
+			recursive=recursive)
+	}
+	attr(ans, "recursive") <- recursive
+	names(ans) <- names(exprs)
+	ans
+}
+
+eval_at <- function(expr, data, i = NULL, j = NULL,
+	split_along = NULL, group = NULL, reduce = "+",
+	recursive = !is.null(split_along))
+{
+	vars <- all.vars(expr)
+	data <- lapply(vars,
+		function(nm) {
+			x <- data[[nm]]
+			if ( !is.null(i) && !is.null(j) ) {
+				if ( is.null(dim(x)) ) {
+					x <- subset_list(x[i], j)
+				} else {
+					x <- x[i,j,drop=FALSE]
+				}
+			} else if ( !is.null(i) ) {
+				if ( is.null(dim(x)) ) {
+					x <- x[i]
+				} else {
+					x <- x[i,,drop=FALSE]
+				}
+			} else if ( !is.null(j) ) {
+				if ( is.null(dim(x)) ) {
+					x <- subset_list(x, j)
+				} else {
+					x <- x[,j,drop=FALSE]
+				}
+			}
+			if ( !is.null(split_along) )
+				x <- array2list(x, split_along)
+			if ( !is.null(group) ) {
+				FUN <- match.fun(reduce)
+				group <- factor(group, levels=unique(group))
+				x <- lapply(levels(group),
+					function(g) Reduce(FUN, x[group %in% g]))
+				names(x) <- levels(group)
+			}
+			x
+		})
+	if ( recursive ) {
+		EVAL <- function(...)
+		{
+			datalist <- list(...)
+			names(datalist) <- vars
+			eval(expr, envir=datalist)
+		}
+		do.call(Map, c(list(EVAL), data))
+	} else {
+		names(data) <- vars
+		eval(expr, envir=data)
+	}
+}
+
