@@ -2,7 +2,7 @@
 #### Define visual encoding classes for 'vizi' ####
 ## ------------------------------------------------
 
-vizi <- function(data, ..., encoding = NULL, params = NULL)
+vizi <- function(data, ..., encoding = NULL, mark = NULL, params = NULL)
 {
 	if ( missing(data) )
 		data <- NULL
@@ -13,16 +13,21 @@ vizi <- function(data, ..., encoding = NULL, params = NULL)
 	coord <- list(xlim=NULL, ylim=NULL, log="", asp=NA, grid=TRUE)
 	plot <- structure(list(encoding=encoding, channels=channels,
 		marks=list(), coord=coord, params=params), class="vizi_plot")
+	if ( !is.null(mark) ) {
+		for ( mk in mark )
+			plot <- add_mark(plot, mk)
+	}
+	plot
 }
 
 add_mark <- function(plot, mark, ..., encoding = NULL,
 	data = NULL, trans = NULL, params = NULL)
 {
 	if ( !inherits(plot, c("vizi_plot", "vizi_facets")) )
-		stop("'plot' must inherit from 'vizi_plot' or 'vizi_facets")
+		matter_error("'plot' must inherit from 'vizi_plot' or 'vizi_facets")
 	cls <- paste0("vizi_", mark)
 	if ( !existsMethod("plot", cls) )
-		stop("no known plot() method for class: ", sQuote(cls))
+		matter_error("no known plot() method for class: ", sQuote(cls))
 	# encode new variables
 	if ( ...length() > 0L || !is.null(encoding) ) {
 		encoding <- merge_encoding(encoding, as_encoding(...))
@@ -44,12 +49,12 @@ add_mark <- function(plot, mark, ..., encoding = NULL,
 				mks[[i]]$encoding <- e
 			}
 			pmks <- plot$plots[[i]]$marks
-			mk <- setNames(list(mks[[i]]), mark)
+			mk <- set_names(list(mks[[i]]), mark)
 			plot$plots[[i]]$marks <- c(pmks, mk)
 		}
 	} else {
 		# assign mark
-		mk <- setNames(list(mk), mark)
+		mk <- set_names(list(mk), mark)
 		plot$marks <- c(plot$marks, mk)
 	}
 	plot
@@ -59,7 +64,7 @@ add_facets <- function(plot, by = NULL, data = NULL,
 	nrow = NA, ncol = NA, labels = NULL, drop = TRUE, free = "")
 {
 	if ( !inherits(plot, c("vizi_plot", "vizi_facets")) )
-		stop("'plot' must inherit from 'vizi_plot' or 'vizi_facets")
+		matter_error("'plot' must inherit from 'vizi_plot' or 'vizi_facets")
 	# encode faceting variable
 	if ( is(by, "formula") ) {
 		by <- parse_formula(by, data)
@@ -100,7 +105,7 @@ as_facets <- function(plotlist, ..., nrow = NA, ncol = NA,
 	all_plots <- all(vapply(plotlist, is, logical(1L), "vizi_plot"))
 	all_facets <- all(vapply(plotlist, is, logical(1L), "vizi_facets"))
 	if ( !all_plots && !all_facets )
-		stop("all plots must inherit from one of 'vizi_plot' or 'vizi_facets")
+		matter_error("all plots must inherit from one of 'vizi_plot' or 'vizi_facets")
 	if ( is.null(labels) ) {
 		labels <- names(plotlist)
 	} else {
@@ -147,7 +152,7 @@ merge_facets <- function(plot1, plot2)
 	if ( !(is(plot1, "vizi_plot") && is(plot2, "vizi_plot")) &&
 		!(is(plot1, "vizi_facets") && is(plot2, "vizi_facets")) )
 	{
-		stop("all plots must inherit from one of 'vizi_plot' or 'vizi_facets")
+		matter_error("all plots must inherit from one of 'vizi_plot' or 'vizi_facets")
 	}
 	# merge channels
 	channels <- merge_channels(plot1$channels, plot2$channels)
@@ -181,6 +186,17 @@ merge_facets <- function(plot1, plot2)
 	}
 }
 
+as_layers <- function(plotlist, ...)
+{
+	# create plot list
+	if ( missing(plotlist) || is.null(plotlist) )
+		plotlist <- list()
+	if ( inherits(plotlist, c("vizi_plot", "vizi_facets")) )
+		plotlist <- list(plotlist)
+	plotlist <- c(plotlist, list(...))
+	do.call(combine, plotlist)
+}
+
 set_title <- function(plot, title)
 {
 	plot$title <- as.character(title)
@@ -246,6 +262,15 @@ set_engine <- function(plot, engine = c("base", "plotly"))
 	plot
 }
 
+as_plotly <- function(plot, source = "A")
+{
+	plot$engineopts <- list(
+		source=source,
+		render=FALSE)
+	output <- plot(plot, engine="plotly")
+	output$engine$plotly
+}
+
 # register for S4 methods
 
 setOldClass("vizi_plot")
@@ -301,8 +326,13 @@ plot_init <- function(plot = NULL, ..., more = list(), n = 1L)
 	if ( e$name == "plotly" )
 	{
 		# initialize plotly
-		if ( is.null(e$plotly) )
-			plot$engine$plotly <- plotly::plot_ly()
+		if ( is.null(e$plotly) ) {
+			if ( is.character(plot$engineopts$source) ) {
+				plot$engine$plotly <- plotly::plot_ly(source=plot$engineopts$source)
+			} else {
+				plot$engine$plotly <- plotly::plot_ly()
+			}
+		}
 		# setup axes
 		xlab <- plot$channels$x$label
 		ylab <- plot$channels$y$label
@@ -356,13 +386,14 @@ plot_init <- function(plot = NULL, ..., more = list(), n = 1L)
 				args$ylim <- rev(args$ylim)
 		}
 		# filter for valid graphical parameters
-		pars <- names(args)
-		pars <- pars[pars %in% names(par(no.readonly=TRUE))]
-		pars <- union(c("xlim", "ylim", "log", "asp"), pars)
-		args <- args[pars]
+		winargs <- c("xlim", "ylim", "log", "asp")
+		pars <- setdiff(names(par(no.readonly=TRUE)), winargs)
+		winargs <- args[intersect(names(args), winargs)]
+		pars <- args[intersect(names(args), pars)]
 		# initialize the 2d plot
 		plot.new()
-		do.call(plot.window, args)
+		do.call(plot.window, winargs)
+		par(pars)
 		# add annotations
 		if ( has_free_x(plot) || is_bottom_panel(n) ) {
 			xl <- plot$channels$x$limits
@@ -388,6 +419,8 @@ plot_init <- function(plot = NULL, ..., more = list(), n = 1L)
 			args$zlim <- plot$coord$zlim
 		if ( is.null(args$zlim) )
 			args$zlim <- plot$channels$z$limits
+		if ( length(unique(args$zlim)) <= 1L )
+			args$zlim <- c(-1, 1) + args$zlim
 		# get x/y/z labels
 		if ( is.null(args$xlab) )
 			args$xlab <- plot$channels$x$label
@@ -484,7 +517,7 @@ plot.vizi_plot <- function(x, add = FALSE, ..., engine = NULL)
 	if ( x$engine$name == "plotly" )
 	{
 		if ( !requireNamespace("plotly") )
-			stop("failed to load required package 'plotly'")
+			matter_error("failed to load required package 'plotly'")
 		if ( !add )
 			plot_init(x, more=par_update(x$params, ...))
 	}
@@ -522,7 +555,8 @@ plot.vizi_plot <- function(x, add = FALSE, ..., engine = NULL)
 			if ( !is.null(x$title) )
 				x$engine$plotly <- plotly::layout(x$engine$plotly,
 					title=list(text=x$title))
-			print(x$engine$plotly)
+			if ( is.null(x$engineopts) || isTRUE(x$engineopts$render) )
+				print(x$engine$plotly)
 		}
 	}
 	x$keys <- keys
@@ -560,7 +594,7 @@ plot.vizi_facets <- function(x, add = FALSE, ..., engine = NULL)
 	if ( x$engine$name == "plotly" )
 	{
 		if ( !requireNamespace("plotly") )
-			stop("failed to load required package 'plotly'")
+			matter_error("failed to load required package 'plotly'")
 		if ( is.null(x$engine$facets) )
 			x$engine$facets <- vector("list", length=n)
 	}
@@ -592,6 +626,17 @@ plot.vizi_facets <- function(x, add = FALSE, ..., engine = NULL)
 			}
 			params <- par_update(x$params, ...)
 			plot_init(x, xlim=xlim, ylim=ylim, more=params, n=n)
+			# fix subplot aspect ratio
+			if ( x$engine$name == "plotly" && !is.na(x$coord$asp) )
+			{
+				if ( i == 1L ) {
+					xanchor <- "x"
+				} else {
+					xanchor <- paste0("x", i)
+				}
+				yaxis <- list(scaleanchor=xanchor, scaleratio=plot$coord$asp)
+				x$engine$plotly <- plotly::layout(x$engine$plotly, yaxis=yaxis)
+			}
 		}
 		keys[[i]] <- plot(plot, add=TRUE, ...)$keys
 		# add facet annotations
@@ -601,12 +646,11 @@ plot.vizi_facets <- function(x, add = FALSE, ..., engine = NULL)
 				mtext(x$labels[i], cex=par("cex"), col=par("col.lab"))
 			if ( x$engine$name == "plotly" )
 			{
-				x$engine$plotly <- plotly::add_annotations(x$engine$plotly,
+				x$engine$facets[[i]] <- plotly::add_annotations(x$engine$plotly,
 					x=0.5, y=1, xanchor="center", yanchor="top",
 					xref="paper", yref="paper", showarrow=FALSE,
 					text=x$labels[i])
-				x$engine$facets[[i]] <- plot$engine$plotly
-				x$engine$plotly <- NULL
+				x$engine$plotly <- NULL # reset for next facet
 			}
 		}
 		if ( x$engine$name == "base" && add )
@@ -640,11 +684,13 @@ plot.vizi_facets <- function(x, add = FALSE, ..., engine = NULL)
 		}
 		if ( x$engine$name == "plotly" )
 		{
+			x$engine$plotly <- plotly::subplot(x$engine$facets,
+				nrows=x$dim[1L], shareX=!has_free_x(x), shareY=!has_free_y(x))
 			if ( !is.null(x$title) )
 				x$engine$plotly <- plotly::layout(x$engine$plotly,
 					title=list(text=x$title))
-			print(plotly::subplot(x$engine$facets, nrows=x$dim[1L],
-				shareX=!has_free_x(x), shareY=!has_free_y(x)))
+			if ( is.null(x$engineopts) || isTRUE(x$engineopts$render) )
+				print(x$engine$plotly)
 		}
 	}
 	x$keys <- keys
@@ -693,6 +739,10 @@ setMethod("plot", "vizi_colorkey", plot.vizi_colorkey)
 vizi_par <- function(..., style = getOption("matter.vizi.style"))
 {
 	params <- getOption("matter.vizi.par")
+	if ( !is.null(style) ) {
+		p <- par_style(tolower(style), new=TRUE)
+		params <- par_update(params, more=p)
+	}
 	args <- list(...)
 	if ( length(args) > 0L ) {
 		if ( length(args) == 1L ) {
@@ -712,10 +762,6 @@ vizi_par <- function(..., style = getOption("matter.vizi.style"))
 	} else {
 		args < names(params)
 	}
-	if ( !is.null(style) ) {
-		p <- par_style(tolower(style))
-		params <- par_update(p, more=params)
-	}
 	if ( length(args) > 0L )
 		params <- params[args]
 	if ( length(params) == 1L )
@@ -726,21 +772,21 @@ vizi_par <- function(..., style = getOption("matter.vizi.style"))
 vizi_style <- function(style = "light", dpal = "Tableau 10", cpal = "Viridis")
 {
 	if ( !missing(style) ) {
-		style <- match.arg(style, c("light", "dark", "transparent"))
+		style <- match.arg(style, names(par_style_options()))
 		options(matter.vizi.style=style)
 	} else {
 		style <- getOption("matter.vizi.style")
 	}
 	if ( !missing(dpal) ) {
 		tryCatch(dpal(dpal)(1L), error=function(e)
-			stop("palette must be one of ", paste0(palette.pals(), collapse=", ")))
+			matter_error("palette must be one of ", paste0(palette.pals(), collapse=", ")))
 		options(matter.vizi.dpal=dpal)
 	} else {
 		dpal <- getOption("matter.vizi.dpal")
 	}
 	if ( !missing(cpal) ) {
 		tryCatch(cpal(cpal)(1L), error=function(e)
-			stop("palette must be one of ", paste0(hcl.pals(), collapse=", ")))
+			matter_error("palette must be one of ", paste0(hcl.pals(), collapse=", ")))
 		options(matter.vizi.cpal=cpal)
 	} else {
 		cpal <- getOption("matter.vizi.cpal")
@@ -791,7 +837,7 @@ as_encoding <- function(x, y, ..., env = NULL)
 normalize_encoding <- function(e)
 {
 	if ( length(e) > 0L ) {
-		setNames(e, to_vizi_name(names(e)))
+		set_names(e, to_vizi_name(names(e)))
 	} else {
 		NULL
 	}
@@ -867,6 +913,8 @@ merge_channels <- function(c1, c2, ...)
 
 merge_limits <- function(l1, l2, ...)
 {
+	if ( missing(l2) )
+		return(l1)
 	if ( ...length() > 0L )
 		l2 <- do.call(merge_limits, list(l2, ...))
 	if ( is_discrete(l1) && is_discrete(l2) ) {
@@ -879,7 +927,7 @@ merge_limits <- function(l1, l2, ...)
 		} else if ( all(is.na(l2)) ) {
 			l1
 		} else {
-			stop("can't merge continuous and discrete channels")
+			matter_error("can't merge continuous and discrete channels")
 		}
 	}
 }
@@ -941,7 +989,7 @@ compute_subscripts <- function(by, nshingles = 6L)
 			v <- as.factor(v)
 			nms <- levels(v)
 			v <- lapply(levels(v), function(lvl) which(v == lvl))
-			setNames(v, nms)
+			set_names(v, nms)
 		} else {
 			shingles(v, breaks=nshingles)
 		}
@@ -964,7 +1012,7 @@ compute_facets <- function(plot, by, nshingles = 6L)
 {
 	n <- unique(lengths(by))
 	if ( length(n) > 1L )
-		stop("'by' has differing numbers of observations")
+		matter_error("'by' has differing numbers of observations")
 	subscripts <- compute_subscripts(by, nshingles)
 	if ( length(subscripts) == 1L ) {
 		dim <- panel_dim_n(prod(lengths(subscripts)))
@@ -976,7 +1024,7 @@ compute_facets <- function(plot, by, nshingles = 6L)
 	ffac <- function(v, p) {
 		fsub <- function(x) {
 			if ( length(x) != n ) {
-				stop("faceting expected ", n, " observations ",
+				matter_error("faceting expected ", n, " observations ",
 					"but encoding has ", length(x), "observations")
 			} else {
 				x[v]
@@ -1009,7 +1057,7 @@ compute_groups <- function(plot, encoding, names)
 {
 	names <- names[names %in% names(encoding)]
 	names <- names[!duplicated(encoding[names])]
-	names <- setNames(names, names)
+	names <- set_names(names, names)
 	groups <- lapply(plot$channels[names],
 		function(ch) {
 			if ( is_discrete(ch$limits) ) {
@@ -1067,7 +1115,7 @@ get_var <- function(x, data)
 {
 	if ( is(x, "formula") ) {
 		if ( length(x) != 2L )
-			stop("formula encodings can only have rhs")
+			matter_error("formula encodings can only have rhs")
 		eval(x[[2L]], envir=data, enclos=environment(x))
 	} else {
 		force(x)
@@ -1114,10 +1162,7 @@ encode_var <- function(name, encoding = NULL,
 		}
 		# subscripts
 		if ( !is.null(subscripts) )
-		{
-			e <- rep_len(e, length(subscripts))
 			e <- e[subscripts]
-		}
 	}
 	e
 }
@@ -1197,11 +1242,11 @@ get_discrete_scheme <- function(channel)
 		shape = seq_fun(14),
 		color = discrete_pal,
 		fill = discrete_pal,
-		alpha = stop(msg),
-		size = stop(msg),
-		linewidth = stop(msg),
+		alpha = matter_error(msg),
+		size = matter_error(msg),
+		linewidth = matter_error(msg),
 		linetype = seq_fun(6),
-		stop(msg))
+		matter_error(msg))
 }
 
 get_continuous_scheme <- function(channel)
@@ -1212,14 +1257,14 @@ get_continuous_scheme <- function(channel)
 		y = , ymin = , ymax = ,
 		z = ,
 		text = NULL,
-		shape = stop(msg),
+		shape = matter_error(msg),
 		color = continuous_pal,
 		fill = continuous_pal,
 		alpha = range_fun(0, 1),
 		size = range_fun(1, 6),
 		linewidth = range_fun(1, 6),
-		linetype = stop(msg),
-		stop(msg))
+		linetype = matter_error(msg),
+		matter_error(msg))
 }
 
 discrete_pal <- function(n)
@@ -1232,8 +1277,8 @@ continuous_pal <- function(n)
 	hcl.colors(n, getOption("matter.vizi.cpal"))
 }
 
-range_fun <- function(from, to) {
-	function(n) seq.int(from, to, length.out=n)
+range_fun <- function(from, to, pow = 1) {
+	function(n) seq.int(from, to, length.out=n)^pow
 }
 
 seq_fun <- function(max_n) {
@@ -1263,7 +1308,7 @@ shape_pal <- function(n = 20L) {
 		"solid circle" = 19,
 		"bullet" = 20L)
 	if ( n > length(pal) )
-		stop("n [", n, "] too large for shape palette")
+		matter_error("n [", n, "] too large for shape palette")
 	pal[seq_len(n)]
 }
 
@@ -1276,7 +1321,7 @@ line_pal <- function(n = 6L) {
 		"longdash" = 5L,
 		"twodash" = 6L)
 	if ( n > length(pal) )
-		stop("n [", n, "] too large for line palette")
+		matter_error("n [", n, "] too large for line palette")
 	pal[seq_len(n)]
 }
 
@@ -1300,7 +1345,7 @@ encode_scheme <- function(x, scheme, limits)
 		scheme <- rep_len(scheme, nlevels(x))
 	} else {
 		dx <- diff(limits)
-        if ( dx == 0 ) {
+		if ( dx == 0 ) {
 			x <- rep.int(1L, length(x))
 		} else {
 			breaks <- seq.int(limits[1L], limits[2L], length.out=n + 1L)
@@ -1309,7 +1354,7 @@ encode_scheme <- function(x, scheme, limits)
 			x <- cut.default(x, breaks=breaks)
 		}
 	}
-	setNames(scheme[as.integer(x)], x)
+	set_names(scheme[as.integer(x)], x)
 }
 
 encode_legends <- function(channels, params, type = NULL)
